package org.grails.plugins.springsecurity.service.acl

import org.apache.log4j.Level
import org.apache.log4j.Logger

import org.grails.plugins.springsecurity.acl.AclClass
import org.grails.plugins.springsecurity.acl.AclEntry
import org.grails.plugins.springsecurity.acl.AclObjectIdentity
import org.grails.plugins.springsecurity.acl.AclSid

import org.springframework.security.Authentication
import org.springframework.security.acls.AccessControlEntry
import org.springframework.security.acls.Acl
import org.springframework.security.acls.AlreadyExistsException
import org.springframework.security.acls.ChildrenExistException
import org.springframework.security.acls.MutableAcl
import org.springframework.security.acls.MutableAclService
import org.springframework.security.acls.NotFoundException
import org.springframework.security.acls.domain.AccessControlEntryImpl
import org.springframework.security.acls.objectidentity.ObjectIdentity
import org.springframework.security.acls.objectidentity.ObjectIdentityImpl
import org.springframework.security.acls.sid.GrantedAuthoritySid
import org.springframework.security.acls.sid.PrincipalSid
import org.springframework.security.acls.sid.Sid
import org.springframework.security.context.SecurityContextHolder as SCH
import org.springframework.util.Assert

class AclService implements MutableAclService {

	private final Logger log = Logger.getLogger(getClass())

	static transactional = true

	def aclLookupStrategy
	def aclCache
	def messageSource

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.MutableAclService#createAcl(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity)
	 */
	MutableAcl createAcl(ObjectIdentity objectIdentity) throws AlreadyExistsException {
		Assert.notNull objectIdentity, 'Object Identity required'

		// Check this object identity hasn't already been persisted
		if (retrieveObjectIdentity(objectIdentity)) {
			throw new AlreadyExistsException("Object identity '$objectIdentity' already exists")
		}

		// Need to retrieve the current principal, in order to know who "owns" this ACL (can be changed later on)
		PrincipalSid sid = new PrincipalSid(SCH.context.authentication)

		// Create the acl_object_identity row
		createObjectIdentity objectIdentity, sid

		// TODO session flush(), clear()?
		return readAclById(objectIdentity)
	}

	protected void createObjectIdentity(ObjectIdentity object, Sid owner) {
		AclSid ownerSid = createOrRetrieveSid(owner, true)
		AclClass aclClass = createOrRetrieveClass(object.javaType, true)
		save new AclObjectIdentity(
				aclClass: aclClass,
				objectId: object.identifier,
				owner: ownerSid,
				entriesInheriting: true)
	}

	protected AclSid createOrRetrieveSid(Sid sid, boolean allowCreate) {
		Assert.notNull sid, 'Sid required'

		String sidName
		boolean principal
		if (sid instanceof PrincipalSid) {
			sidName = sid.principal
			principal = true
		}
		else if (sid instanceof GrantedAuthoritySid) {
			sidName = sid.grantedAuthority
			principal = false
		}
		else {
			throw new IllegalArgumentException('Unsupported implementation of Sid')
		}

		AclSid aclSid = AclSid.findBySidAndPrincipal(sidName, principal)
		if (!aclSid && allowCreate) {
			aclSid = save(new AclSid(sid: sidName, principal: principal))
		}
		return aclSid
	}

	protected AclClass createOrRetrieveClass(Class<?> clazz, boolean allowCreate) {
		AclClass aclClass = AclClass.findByClassName(clazz.name)
		if (!aclClass && allowCreate) {
			aclClass = save(new AclClass(className: clazz.name))
		}
		return aclClass
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.MutableAclService#deleteAcl(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity, boolean)
	 */
	void deleteAcl(ObjectIdentity objectIdentity, boolean deleteChildren) throws ChildrenExistException {

		Assert.notNull objectIdentity, 'Object Identity required'
		Assert.notNull objectIdentity.identifier, "Object Identity doesn't provide an identifier"

		if (deleteChildren) {
			for (child in findChildren(objectIdentity)) {
				deleteAcl(child, true)
			}
		}

		AclObjectIdentity oid = retrieveObjectIdentity(objectIdentity)

		// Delete this ACL's ACEs in the acl_entry table
		deleteEntries oid

		// Delete this ACL's acl_object_identity row
		oid.delete()

		// Clear the cache
		aclCache.evictFromCache(objectIdentity)
	}

	protected void deleteEntries(AclObjectIdentity oid) {
		AclEntry.executeUpdate(
				"DELETE FROM AclEntry ae " +
				"WHERE ae.aclObjectIdentity = ?", [oid])
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.MutableAclService#updateAcl(
	 * 	org.springframework.security.acls.MutableAcl)
	 */
	MutableAcl updateAcl(MutableAcl acl) throws NotFoundException {
		Assert.notNull acl.id, "Object Identity doesn't provide an identifier"

		// Delete this ACL's ACEs in the acl_entry table
		deleteEntries retrieveObjectIdentity(acl.objectIdentity)

		// Create this ACL's ACEs in the acl_entry table
		createEntries acl

		// Change the mutable columns in acl_object_identity
		updateObjectIdentity acl

		// Clear the cache, including children
		clearCacheIncludingChildren acl.objectIdentity

		// TODO session flush(), clear()?
		return readAclById(acl.objectIdentity)
	}

	protected void createEntries(MutableAcl acl) {
		int i = 0
		for (AccessControlEntry entry in acl.entries) {
			Assert.isInstanceOf AccessControlEntryImpl, entry, 'Unknown ACE class'
			save new AclEntry(
					aclObjectIdentity: AclObjectIdentity.get(acl.id),
					aceOrder: i++,
					sid: createOrRetrieveSid(entry.sid, true),
					mask: entry.permission.mask,
					granting: entry.isGranting(),
					auditSuccess: entry.isAuditSuccess(),
					auditFailure: entry.isAuditFailure())
		}
	}

	protected void updateObjectIdentity(MutableAcl acl) {
		Assert.notNull acl.owner, "Owner is required in this implementation"

		AclObjectIdentity aclObjectIdentity = AclObjectIdentity.get(acl.id)

		AclObjectIdentity parent
		if (acl.parentAcl) {
			def oii = acl.parentAcl.objectIdentity
			Assert.isInstanceOf ObjectIdentityImpl, oii,
				'Implementation only supports ObjectIdentityImpl'
			parent = retrieveObjectIdentity(oii)
		}

		aclObjectIdentity.parent = parent
		aclObjectIdentity.owner = createOrRetrieveSid(acl.owner, true)
		aclObjectIdentity.entriesInheriting = acl.isEntriesInheriting()
	}

	protected void clearCacheIncludingChildren(ObjectIdentity objectIdentity) {
		Assert.notNull objectIdentity, 'ObjectIdentity required'
		for (child in findChildren(objectIdentity)) {
			clearCacheIncludingChildren child
		}
		aclCache.evictFromCache(objectIdentity)
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.AclService#findChildren(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity)
	 */
	ObjectIdentity[] findChildren(ObjectIdentity parent) {
		def children = AclObjectIdentity.executeQuery(
				"FROM AclObjectIdentity " +
				"WHERE parent.objectId = :objectId " +
				"  AND parent.aclClass.className = :className",
				[objectId: parent.identifier,
				 className: parent.javaType.name])

		if (!children) {
			return null
		}

		def results = []
		for (AclObjectIdentity aclObjectIdentity in children) {
			results << new ObjectIdentityImpl(
					lookupClass(aclObjectIdentity.aclClass.className), aclObjectIdentity.objectId)
		}

		return results as ObjectIdentity[]
	}

	protected Class<?> lookupClass(String className) {
		// workaround for Class.forName() not working in tests
		return Class.forName(className, true, Thread.currentThread().contextClassLoader)
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.AclService#readAclById(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity)
	 */
	Acl readAclById(ObjectIdentity object) throws NotFoundException {
		return readAclById(object, null)
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.AclService#readAclById(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity,
	 * 	org.springframework.security.acls.sid.Sid[])
	 */
	Acl readAclById(ObjectIdentity object, Sid[] sids) throws NotFoundException {
		Map<ObjectIdentity, Acl> map = readAclsById(object as ObjectIdentity[], sids)
		Assert.isTrue map.containsKey(object),
				"There should have been an Acl entry for ObjectIdentity $object"
		return map[object]
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.AclService#readAclsById(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity[])
	 */
	Map<ObjectIdentity, Acl> readAclsById(ObjectIdentity[] objects) throws NotFoundException {
		return readAclsById(objects, null)
	}

	/**
	 * {@inheritDoc}
	 * @see org.springframework.security.acls.AclService#readAclsById(
	 * 	org.springframework.security.acls.objectidentity.ObjectIdentity[],
	 * 	org.springframework.security.acls.sid.Sid[])
	 */
	Map<ObjectIdentity, Acl> readAclsById(ObjectIdentity[] objects, Sid[] sids) throws NotFoundException {
		Map<ObjectIdentity, Acl> result = aclLookupStrategy.readAclsById(objects, sids)
		// Check every requested object identity was found (throw NotFoundException if needed)
		for (object in objects) {
			if (!result.containsKey(object)) {
				throw new NotFoundException(
						"Unable to find ACL information for object identity '$object'")
			}
		}
		return result
	}

	protected AclObjectIdentity retrieveObjectIdentity(ObjectIdentity oid) {
		return AclObjectIdentity.executeQuery(
				"FROM AclObjectIdentity " +
				"WHERE aclClass.className = :className " +
				"  AND objectId = :objectId",
				[className: oid.javaType.name,
				 objectId: oid.identifier])[0]
	}

	protected save(bean) {
		bean.validate()
		if (bean.hasErrors()) {
			if (log.isEnabledFor(Level.WARN)) {
				def message = new StringBuilder(
						"problem creating ${bean.getClass().simpleName}: $bean")
				def locale = Locale.getDefault()
				for (fieldErrors in bean.errors) {
					for (error in fieldErrors.allErrors) {
						message.append('\n\t')
						message.append(messageSource.getMessage(error, locale))
					}
				}
				log.warn message
			}
			bean.discard()
		}
		else {
			bean.save()
		}

		return bean
	}
}
