package org.grails.plugins.springsecurity.service.acl

import org.springframework.security.Authentication
import org.springframework.security.acls.Acl
import org.springframework.security.acls.MutableAcl
import org.springframework.security.acls.NotFoundException
import org.springframework.security.acls.Permission
import org.springframework.security.acls.objectidentity.ObjectIdentity
import org.springframework.security.acls.objectidentity.ObjectIdentityImpl

import org.springframework.security.acls.sid.GrantedAuthoritySid
import org.springframework.security.acls.sid.PrincipalSid
import org.springframework.security.acls.sid.Sid
import org.springframework.security.acls.sid.SidRetrievalStrategy
import org.springframework.security.acls.sid.SidRetrievalStrategyImpl

class AclUtilService {

	private SidRetrievalStrategy sidRetrievalStrategy = new SidRetrievalStrategyImpl()

	def aclService
	def objectIdentityRetrievalStrategy

	void addPermission(object, recipient, Permission permission) {
		addPermission(object.getClass(), object.id, recipient, permission)
	}

	void addPermission(Class<?> domainClass, long id, recipient, Permission permission) {

		Sid sid = createSid(recipient)

		ObjectIdentity oid = new ObjectIdentityImpl(domainClass, id)
		MutableAcl acl
		try {
			acl = aclService.readAclById(oid)
		}
		catch (NotFoundException e) {
			acl = aclService.createAcl(oid)
		}

		acl.insertAce acl.entries.length, permission, sid, true
		aclService.updateAcl acl

		log.debug "Added permission $permission for Sid $sid for $domainClass.name with id $id"
	}

	void deletePermission(object, recipient, Permission permission) {
		deletePermission(object.getClass(), object.id, recipient, permission)
	}

	void deletePermission(Class<?> domainClass, long id, recipient, Permission permission) {
		Sid sid = createSid(recipient)
		ObjectIdentity oid = new ObjectIdentityImpl(domainClass, id)
		MutableAcl acl = aclService.readAclById(oid)

		acl.entries.eachWithIndex { entry, i ->
			if (entry.sid.equals(sid) && entry.permission.equals(permission)) {
				acl.deleteAce(i)
			}
		}

		aclService.updateAcl(acl)

		log.debug "Deleted ${domainClass.name}($id) ACL permissions for recipient $recipient"
	}

	boolean hasPermission(Authentication authentication, domainObject, Permission[] permissions) {
		ObjectIdentity objectIdentity = objectIdentityRetrievalStrategy.getObjectIdentity(domainObject)
		Sid[] sids = sidRetrievalStrategy.getSids(authentication)
		try {
			Acl acl = aclService.readAclById(objectIdentity, sids)
			return acl.isGranted(permissions, sids, false)
		}
		catch (NotFoundException e) {
			return false
		}
	}

	boolean hasPermission(Authentication authentication, domainObject, Permission permission) {
		return hasPermission(authentication, domainObject, [permission] as Permission[])
	}

	private Sid createSid(recipient) {
		if (recipient instanceof String) {
			return recipient.startsWith('ROLE_') ?
					new GrantedAuthoritySid(recipient) :
					new PrincipalSid(recipient)
		}

		if (recipient instanceof Sid) {
			return recipient
		}

		if (recipient instanceof Authentication) {
			return new PrincipalSid(recipient)
		}

		throw new IllegalArgumentException('recipient must be a String, Sid, or Authentication')
	}
}
