/* Copyright 2006-2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.codehaus.groovy.grails.plugins.springsecurity.acl

import java.lang.reflect.Method

import grails.util.GrailsNameUtils

import org.apache.log4j.Logger

import org.springframework.aop.framework.autoproxy.BeanNameAutoProxyCreator

import org.springframework.security.ConfigAttributeDefinition
import org.springframework.security.GrantedAuthority
import org.springframework.security.GrantedAuthorityImpl
import org.springframework.security.SecurityConfig
import org.springframework.security.acls.domain.AclAuthorizationStrategyImpl
import org.springframework.security.acls.domain.BasePermission
import org.springframework.security.acls.domain.ConsoleAuditLogger
import org.springframework.security.acls.jdbc.EhCacheBasedAclCache
import org.springframework.security.acls.jdbc.JdbcMutableAclService
import org.springframework.security.acls.objectidentity.ObjectIdentityRetrievalStrategy
import org.springframework.security.afterinvocation.AclEntryAfterInvocationCollectionFilteringProvider
import org.springframework.security.afterinvocation.AclEntryAfterInvocationProvider
import org.springframework.security.afterinvocation.AfterInvocationProviderManager
import org.springframework.security.intercept.method.MapBasedMethodDefinitionSource
import org.springframework.security.intercept.method.aopalliance.MethodSecurityInterceptor
import org.springframework.security.vote.AclEntryVoter
import org.springframework.security.vote.AffirmativeBased
import org.springframework.util.Assert

import org.codehaus.groovy.grails.commons.GrailsClassUtils as GCU

import org.codehaus.groovy.grails.plugins.springsecurity.Secured
import org.codehaus.groovy.grails.plugins.springsecurity.acl.AclVoter
import org.codehaus.groovy.grails.plugins.springsecurity.acl.AclVoters
import org.codehaus.groovy.grails.plugins.springsecurity.acl.GormObjectIdentityRetrievalStrategy
import org.codehaus.groovy.grails.plugins.springsecurity.acl.NullAclAuditLogger
import org.grails.plugins.springsecurity.service.acl.LookupStrategyService

import org.springframework.cache.ehcache.EhCacheFactoryBean
import org.springframework.cache.ehcache.EhCacheManagerFactoryBean

class AclPluginConfig {
	
	private final Logger log = Logger.getLogger(getClass())

	private static final List<String> NON_SECURABLE_METHODS =
		['invokeMethod', 'getMetaClass', 'setMetaClass', 'getProperty', 'setProperty',
		 'isTransactional', 'getTransactional', 'setTransactional']

	def doWithSpring = { conf ->

		debug 'configuring spring security ACLs ...'

		// create decision voters
		Map voterConfig = buildVoterConfig(conf, application)
		debug "voterConfig: $voterConfig"
		def aclAccessDecisionManagerDecisionVoters = [ref('roleVoter')]
		voterConfig.each { beanName, voterData ->
			"$beanName"(AclEntryVoter, ref('aclService'), voterData.configAttribute, voterData.permissions) {
				processDomainObjectClass = voterData.domainObjectClass
			}
			aclAccessDecisionManagerDecisionVoters << ref(beanName)
			debug "created AclEntryVoter $beanName for domain class $voterData.domainObjectClass.name with configAttribute $voterData.configAttribute and permissions $voterData.permissions"
		}

		aclAccessDecisionManager(AffirmativeBased) {
			allowIfAllAbstainDecisions = false
			decisionVoters = aclAccessDecisionManagerDecisionVoters
		}

		objectIdentityRetrievalStrategy(GormObjectIdentityRetrievalStrategy)

		// acl cache
		aclCacheManager(EhCacheManagerFactoryBean)
		ehcacheAclCache(EhCacheFactoryBean) {
			cacheManager = ref('aclCacheManager')
			cacheName = 'aclCache'
		}
		aclCache(EhCacheBasedAclCache, ehcacheAclCache)

		aclAuthorizationStrategy(AclAuthorizationStrategyImpl,
				[new GrantedAuthorityImpl(conf.acl.authority.changeOwnership),
				 new GrantedAuthorityImpl(conf.acl.authority.modifyAuditingDetails),
				 new GrantedAuthorityImpl(conf.acl.authority.changeAclDetails)] as GrantedAuthority[])

		aclAuditLogger(NullAclAuditLogger)

		// Phillip Merensky's version based on Stephan February's - also requires JdbcMutableAclService below
//		lookupStrategy(LookupStrategyService) {
//			aclCache = ref('aclCache')
//			aclAuthorizationStrategy = ref('aclAuthorizationStrategy')
//		}

//		Activate this and deactivate the lines above to use the default lookup strategy
//		lookupStrategy(org.springframework.security.acls.jdbc.BasicLookupStrategy,
//				ref('dataSource'),
//				ref('aclCache'),
//				ref('aclAuthorizationStrategy'),
//				ref('aclAuditLogger'))

//		debug "Class Identity Query: $conf.acl.aclClassIdentityQuery"
//		aclService(JdbcMutableAclService, ref('dataSource'), ref('lookupStrategy'), ref('aclCache')) {
//			classIdentityQuery = conf.acl.aclClassIdentityQuery
//			sidIdentityQuery = conf.acl.aclSidIdentityQuery
//		}

		aclLookupStrategy(GormAclLookupStrategy) {
			aclCache = ref('aclCache')
			aclAuthorizationStrategy = ref('aclAuthorizationStrategy')
			auditLogger = ref('aclAuditLogger')
		}

		// processes AFTER_ACL_COLLECTION_READ configuration settings;
		// filters out records you don't have access to
		afterAclCollectionRead(AclEntryAfterInvocationCollectionFilteringProvider, ref('aclService'),
				[BasePermission.READ])

		// processes AFTER_ACL_READ configuration settings;
		// determines access to single instances returned
		afterAclRead(AclEntryAfterInvocationProvider, ref('aclService'),
				[BasePermission.READ])

		afterInvocationManager(AfterInvocationProviderManager) {
			providers = [ref('afterAclRead'), ref('afterAclCollectionRead')]
		}

		debug 'Starting to loop through Service classes...'
		// look for the static property 'springSecurityACL' or annotations in each
		// service class and add interceptors for methods defined there
		for (serviceClass in application.serviceClasses) {

			Map springSecurityACL = extractSpringSecurityACL(serviceClass.clazz)
			if (!springSecurityACL) {
				continue
			}

			String className = serviceClass.clazz.name
			debug "configuring service class $className"

			String interceptorObjectDefinitionSource = buildInterceptorObjectDefinitionSource(
					className, springSecurityACL).join('\n')
			debug "\tobjectDefinitionSource is $interceptorObjectDefinitionSource"

			//Builds an AFTER interceptor for the current service class
			"${className}_AfterInterceptor"(MethodSecurityInterceptor) {
				authenticationManager = ref('authenticationManager')
				accessDecisionManager = ref('aclAccessDecisionManager')
				afterInvocationManager = ref('afterInvocationManager')
				objectDefinitionSource = interceptorObjectDefinitionSource
			}
			debug "\tcreated MethodSecurityInterceptor ${className}_AfterInterceptor"

			//We only need this if aclvoters are defined in config
			if (!voterConfig.empty) {
				//Builds an BEFORE interceptor for the current service class
				//This is used for method invocation security
				//the interceptor is used in doWithApplicationContext later
				"${className}_BeforeInterceptor"(MethodSecurityInterceptor) {
					authenticationManager = ref('authenticationManager')
					accessDecisionManager = ref('aclAccessDecisionManager')
					afterInvocationManager = ref('afterInvocationManager')
					objectDefinitionSource = interceptorObjectDefinitionSource
				}
				debug "\tcreated MethodSecurityInterceptor ${className}_BeforeInterceptor"
			}

			String serviceName = GrailsNameUtils.getPropertyNameRepresentation(className)
			// assigns interceptors to service class
			"${className}_SecurityManager"(BeanNameAutoProxyCreator) {
				proxyTargetClass = true
				beanNames = [serviceName]
				interceptorNames = "${className}_AfterInterceptor"
			}
			debug "\tassigned interceptor for service $serviceName"
		}
	}

	private Map buildVoterConfig(conf, application) {
		Map config = conf.acl.voters.clone()

		for (dc in application.domainClasses) {
			for (annotation in findAnnotations(dc.clazz)) {
				def permissions = []
				for (String permissionName in annotation.permissions()) {
					permissions << BasePermission."$permissionName"
				}
				config[annotation.name()] = [configAttribute: annotation.configAttribute(),
				                             domainObjectClass: dc.clazz,
				                             permissions: permissions]
			}
		}
		
		config
	}

	private List<AclVoter> findAnnotations(Class<?> domainClass) {
		List<AclVoter> annotations = []

		AclVoter aclVoterAnnotation = domainClass.getAnnotation(AclVoter)
		if (aclVoterAnnotation) {
			annotations << aclVoterAnnotation
		}

		AclVoters aclVotersAnnotation = domainClass.getAnnotation(AclVoters)
		if (aclVotersAnnotation) {
			annotations.addAll aclVotersAnnotation.value() as List
		}
		
		annotations
	}

	// build up the list of permissions name/value pairs
	// e.g. ReportService.list=ROLE_USER,AFTER_ACL_COLLECTION_READ
	private List buildInterceptorObjectDefinitionSource(String className, Map springSecurityACL) {
		def permArray = []
		springSecurityACL.each { methodName, attributes ->
			def permList = []
			for (attribute in attributes) {
				if (!attribute.startsWith('ACL_')) {
					permList << attribute
					debug "Added perm $attribute"
				}
			}
			if (permList) {
				String rule = "${className}.${methodName}=${permList.join(',')}"
				permArray << rule
				debug "\tadding rule $rule"
			}
		}
		permArray
	}

	def doWithApplicationContext = { ctx, conf ->

		//We only need this if aclvoters are defined in config
		if (!buildVoterConfig(conf, application)) {
			return
		}

		// add interceptors for voters to our service classes
		debug 'Starting with adding of voters...'
		for (serviceClass in application.serviceClasses) {

			Map springSecurityACL = extractSpringSecurityACL(serviceClass.clazz)
			if (!springSecurityACL) {
				continue
			}

			String className = serviceClass.clazz.name
			def permMap = createMethodDefinitionSourceMap(className, springSecurityACL)
			if (permMap) {
				ctx.getBean("${className}_BeforeInterceptor").objectDefinitionSource =
					new MapBasedMethodDefinitionSource(permMap)

				String serviceName = GrailsNameUtils.getPropertyNameRepresentation(className)
				ctx.getBean("${className}_SecurityManager").interceptorNames =
					["${className}_AfterInterceptor",
					 "${className}_BeforeInterceptor"] as String[]

				debug "assigned interceptor for service $serviceName"
			}
		}

		//configure error handling for acls -> already configured for the plugin
		//does not work anyhow
		//        org.codehaus.groovy.grails.web.errors.GrailsExceptionResolver apf = ctx.getBean('exceptionHandler')
		//        def props = new Properties()
		//        props.setProperty('AccessDeniedException', '/index')
		//        apf.setExceptionMappings(props)
	}

	// build up the list of permissions name/value pairs
	// e.g. ReportService.list=ROLE_USER,AFTER_ACL_COLLECTION_READ
	private Map createMethodDefinitionSourceMap(String className, Map springSecurityACL) {
		def permMap = [:]
		springSecurityACL.each { methodName, attributes ->
			def configAttributes = []
			for (attribute in attributes) {
				if (attribute.startsWith('ACL_')) {
					configAttributes << new SecurityConfig(attribute)
					debug "Added perm $attribute"
				}
			}
			if (configAttributes) {
				def configDefinition = new ConfigAttributeDefinition(configAttributes)
				permMap[className + '.' + methodName] = new ConfigAttributeDefinition(configAttributes)
				debug "\tadding ConfigAttributeDefinition $configDefinition for ${className}.${methodName}()"
			}
		}
		permMap
	}

	private Map extractSpringSecurityACL(Class<?> serviceClass) {
		if (GCU.isStaticProperty(serviceClass, 'springSecurityACL')) {
			return serviceClass.springSecurityACL
		}

		def methods = findSecurableMethods(serviceClass.methods)

		Map<String, List<String>> springSecurityACL = [:]
		Secured annotation = serviceClass.getAnnotation(Secured)
		if (annotation) {
			// map all methods with the annotation's values; can be overridden at the method level
			List<String> roleNames = annotation.value() as List
			for (Method method in methods) {
				springSecurityACL[method.name] = roleNames
			}
		}

		for (Method method in methods) {
			annotation = method.getAnnotation(Secured)
			if (annotation) {
				springSecurityACL[method.name] = annotation.value() as List
			}
		}

		springSecurityACL
	}

	private List<Method> findSecurableMethods(Method[] methods) {

		def objectMethods = Object.methods as List

		List<Method> securableMethods = []
		for (Method method in methods) {
			if (!method.name.startsWith('super$') &&
					!objectMethods.contains(method) &&
					!NON_SECURABLE_METHODS.contains(method.name)) {
				securableMethods << method
			}
		}

		securableMethods
	}

	private void debug(message) {
		log.debug message
	}
}
